/*
*/
#define SYSTEM_PRIVATE
#include <exec/memory.h>
#include <exec/tasks.h>
#include <exec/execbase.h>
#include <dos/dosextens.h>
#include <dos/filehandler.h>
#include <dos/datetime.h>
#include <dos/dostags.h>
#include <proto/exec.h>
#include <proto/dos.h>
#include <clib/alib_protos.h>
#include <aros/machine.h>
#include <strings.h>
#include <stddef.h>
#include "doslist_version.h"

static const char version[] = VERSTAG;

#define ARG_TEMPLATE "NAME,ADDRESS/N,DEVICES/S,VOLUMES/S,ASSIGNS/S,VERBOSE/S"

enum
{
	ARG_NAME = 0,
	ARG_ADDRESS,
	ARG_DEVICES,
	ARG_VOLUMES,
	ARG_ASSIGNS,
	ARG_VERBOSE,
	NOOFARGS
};


struct data
{
	struct ExecBase		*SysBase;
	struct DosLibrary	*DOSBase;
	LONG			arg_verbose;
	LONG			arg_quiet;
	struct MinList		printblist;
	IPTR			args[NOOFARGS];
};

#define SysBase		data->SysBase
#define DOSBase		data->DOSBase
#define printblist	data->printblist
#define args		data->args

struct fmtbuffer
{
  struct MinNode node;
  LONG           index;
  UBYTE          buf[0];
};
#define PRINTBBUFSIZE 256

int	ShowLock(struct data	*data,BPTR lock);
int	ShowDevices(struct data	*data);
int	ShowVolumes(struct data	*data);
int	ShowAssigns(struct data	*data);

static
void InitPrintb(struct data *data);
static
void Printb(struct data *data, const char *fmt, ...);
static
void DumpPrintb(struct data *data, BPTR fh);

#define myPrintf(fmt, args...) if (arg_vmem) Printb(data, fmt,##args); else Printf(fmt,##args);
#define myIsInteractive(fh) arg_vmem ? FALSE : IsInteractive(fh)


/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

void *memclr(void *, size_t);

int main(void)
{
struct data _data, *data = &_data;
struct RDArgs *rda;
LONG error=RETURN_FAIL;

	memset(data, 0, sizeof(*data));
	SysBase=*((struct ExecBase**) 4);
	if ((DOSBase=(struct DosLibrary *)OpenLibrary("dos.library",51)))
	{
		rda=(struct RDArgs *)ReadArgs(ARG_TEMPLATE, args, NULL);

		if (rda)
		{
			error=RETURN_OK;


			if (args[ARG_DEVICES] || (!(args[ARG_VOLUMES] || args[ARG_ASSIGNS])))
			{
				error = ShowDevices(data);
			}

			if (!error && (args[ARG_VOLUMES] || (!(args[ARG_DEVICES] || args[ARG_ASSIGNS]))))
			{
				error = ShowVolumes(data);
			}

			if (!error && (args[ARG_ASSIGNS] || (!(args[ARG_DEVICES] || args[ARG_VOLUMES]))))
			{
				error = ShowAssigns(data);
			}

			FreeArgs(rda);
		}
		if (error != RETURN_OK)
		{
			PrintFault(IoErr(), NULL);
		}
		CloseLibrary((struct Library *)DOSBase);
	}

	return error;
}


/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

static struct Process *GetPortProc(struct data *data, const struct MsgPort *MyPort)
{
	struct Process *MyProcess = NULL;

	if (TypeOfMem((APTR)MyPort) /*&& MyPort->mp_Node.ln_Type == NT_MSGPORT*/)
	{
		if ((MyPort->mp_Flags & PF_ACTION) == PA_SIGNAL)
		{
			MyProcess =(struct Process*) MyPort->mp_SigTask;
		}
		else
		{
			MyProcess =(struct Process*) ((UBYTE*) MyProcess - offsetof(struct Process,pr_MsgPort));
		}

		if (!TypeOfMem(MyProcess) ||
		    (MyProcess->pr_Task.tc_Node.ln_Type != NT_TASK &&
		     MyProcess->pr_Task.tc_Node.ln_Type != NT_PROCESS))
		{
			MyProcess = NULL;
		}
	}

	return MyProcess;
}


int ShowDevices(struct data	*data)
{
struct DosList	*dosentry;
int		error=RETURN_OK;

	InitPrintb(data);

	if ((dosentry=AttemptLockDosList(LDF_DEVICES | LDF_READ)))
	{
		while ((dosentry=NextDosEntry(dosentry, LDF_DEVICES)))
		{
			char	*namearg;
			char	namebuf[256];
			struct DosAttrBuffer name={namebuf,sizeof(namebuf)};
			struct DosAttrBuffer value={0,sizeof(ULONG)};

			if ((namearg=(char*) args[ARG_NAME]))
			{
				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DN_Name,&name,
							TAG_DONE))
				{
					if ((stricmp(name.dab_Ptr,namearg)!=0))
					{
						continue;
					}
				}
				else
				{
					continue;
				}
			}

			if (args[ARG_ADDRESS])
			{
				if (*((ULONG*) args[ARG_ADDRESS]) != (ULONG) dosentry)
				{
					continue;
				}
			}

			if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
						ADO_DN_Name,&name,
						TAG_DONE))
			{
				Printb(data,"<%s> DeviceNode 0x%lx\n",
					name.dab_Ptr,
					dosentry);
			}
			else
			{
				Printb(data,"DeviceNode 0x%lx\n",
					dosentry);
			}


			if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
						ADO_DN_MsgPort,&value,
						TAG_DONE))
			{
				struct Process *MyProcess;
				MyProcess = GetPortProc(data, value.dab_Ptr);
				if (MyProcess)
				{
					Printb(data,"  Process 0x%lx <%s>\n",
						MyProcess,
						MyProcess->pr_Task.tc_Node.ln_Name);
				}
				else
				{
					Printb(data,"  Not Mounted\n");
				}
			}


			if (args[ARG_VERBOSE])
			{
				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DN_Handler,&name,
							TAG_DONE))
				{
					if (((char*)name.dab_Ptr)[0])
					{
						Printb(data,"  Handler <%s>\n",
							name.dab_Ptr);
					}
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DN_StackSize,&value,
							TAG_DONE))
				{
					Printb(data,"  StackSize %ld\n",
							value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DN_Priority,&value,
							TAG_DONE))
				{
					Printb(data,"  Priority %ld\n",
							value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DN_Seglist,&value,
							TAG_DONE))
				{
					Printb(data,"  SegList 0x%lx\n",
							value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DN_GlobalVec,&value,
							TAG_DONE))
				{
					Printb(data,"  GlobalVec %ld\n",
							value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DN_SerialID,&name,
							TAG_DONE))
				{
					if (((char*)name.dab_Ptr)[0])
					{
						Printb(data,"  SerialID <%s>\n",
							name.dab_Ptr);
					}
				}


				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_FSSM,&value,
							TAG_DONE))
				{
					Printb(data,"  FileSysStartupMsg 0x%lx\n",
							value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_FSSM_Device,&name,
							TAG_DONE))
				{
					struct DosAttrBuffer unit={0,sizeof(ULONG)};

					if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
								ADO_FSSM_Unit,&unit,
								TAG_DONE))
					{
						struct DosAttrBuffer flags={0,sizeof(ULONG)};
						if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
									ADO_FSSM_Flags,&flags,
									TAG_DONE))
						{
							Printb(data,"    <%s:%ld> Flags 0x%lx\n",
								name.dab_Ptr,
								unit.dab_Ptr,
								flags.dab_Ptr);
						}
					}
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE,&value,
							TAG_DONE))
				{
					Printb(data,"  Environment 0x%lx\n",
						value.dab_Ptr);
				}


				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_SizeBlock,&value,
							TAG_DONE))
				{
					Printb(data,"    SizeBlock %ld\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_SecOrg,&value,
							TAG_DONE))
				{
					Printb(data,"    SecOrg %ld\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_NumHeads,&value,
							TAG_DONE))
				{
					Printb(data,"    Surfaces %ld\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_SecsPerBlk,&value,
							TAG_DONE))
				{
					Printb(data,"    SectorsPerBlock %ld\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_BlksPerTrack,&value,
							TAG_DONE))
				{
					Printb(data,"    BlocksPerTrack %ld\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_PreFac,&value,
							TAG_DONE))
				{
					Printb(data,"    Reserved %ld\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_PreAlloc,&value,
							TAG_DONE))
				{
					Printb(data,"    PreAlloc %ld\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_Interleave,&value,
							TAG_DONE))
				{
					Printb(data,"    Interleave %ld\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_LowCyl,&value,
							TAG_DONE))
				{
					Printb(data,"    LowCyl %ld\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_HighCyl,&value,
							TAG_DONE))
				{
					Printb(data,"    HighCyl %ld\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_NumBuffers,&value,
							TAG_DONE))
				{
					Printb(data,"    NumBuffers %ld\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_BufMemType,&value,
							TAG_DONE))
				{
					Printb(data,"    BufMemType %ld\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_MaxTransfer,&value,
							TAG_DONE))
				{
					Printb(data,"    MaxTransfer 0x%lx\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_Mask,&value,
							TAG_DONE))
				{
					Printb(data,"    Mask 0x%lx\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_BootPri,&value,
							TAG_DONE))
				{
					Printb(data,"    BootPri %ld\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_DosType,&value,
							TAG_DONE))
				{
					Printb(data,"    DosType 0x%lx\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_Baud,&value,
							TAG_DONE))
				{
					Printb(data,"    Baud %ld\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_BootBlocks,&value,
							TAG_DONE))
				{
					Printb(data,"    BootBlocks %ld\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_Control,&name,
							TAG_DONE))
				{
					if (((char*)name.dab_Ptr)[0])
					{
						Printb(data,"    Control <%s>\n",
							name.dab_Ptr);
					}
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DE_TableSize,&value,
							TAG_DONE))
				{
					Printb(data,"    TableSize %ld\n",
						value.dab_Ptr);
				}

				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DN_Startup,&name,
							TAG_DONE))
				{
					if (((char*)name.dab_Ptr)[0])
					{
						Printb(data,"  Startup <%s>\n",
							name.dab_Ptr);
					}
				}
				if (GetDosObjectAttrTags(DOS_DEVICENODE,dosentry,
							ADO_DN_StartupValue,&value,
							TAG_DONE))
				{
					Printb(data,"  Startup %ld\n",
						value.dab_Ptr);
				}
			}

			if(SetSignal(0L,SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C)
			{
				error = RETURN_FAIL;
				SetIoErr(ERROR_BREAK);
				break;
			}
		}
		UnLockDosList(LDF_DEVICES | LDF_READ);
		DumpPrintb(data, Output());
	}
	return(error);
}


/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

int ShowVolumes(struct data	*data)
{
struct DosList	*dosentry;
int		error=RETURN_OK;

	InitPrintb(data);

	if ((dosentry=AttemptLockDosList(LDF_VOLUMES | LDF_READ)))
	{
		while ((dosentry=NextDosEntry(dosentry, LDF_VOLUMES)))
		{
			char	*namearg;
			char	namebuf[256];
			struct DosAttrBuffer name={namebuf,sizeof(namebuf)};
			struct DosAttrBuffer value={0,sizeof(ULONG)};

			if ((namearg=(char*) args[ARG_NAME]))
			{
				if (GetDosObjectAttrTags(DOS_VOLUMENODE,dosentry,
							ADO_VN_Name,&name,
							TAG_DONE))
				{
					if ((stricmp(name.dab_Ptr,namearg)!=0))
					{
						continue;
					}
				}
				else
				{
					continue;
				}
			}

			if (args[ARG_ADDRESS])
			{
				if (*((ULONG*) args[ARG_ADDRESS]) != (ULONG) dosentry)
				{
					continue;
				}
			}

			if (GetDosObjectAttrTags(DOS_VOLUMENODE,dosentry,
						ADO_VN_Name,&name,
						TAG_DONE))
			{
				Printb(data,"<%s> DeviceNode 0x%lx\n",
					name.dab_Ptr,
					dosentry);
			}
			else
			{
				Printb(data,"DeviceNode 0x%lx\n",
					dosentry);
			}

			if (GetDosObjectAttrTags(DOS_VOLUMENODE,dosentry,
						ADO_DN_MsgPort,&value,
						TAG_DONE))
			{
				struct Process *MyProcess;
				MyProcess = GetPortProc(data, value.dab_Ptr);
				if (MyProcess)
				{
					Printb(data,"  Process 0x%lx <%s>\n",
						MyProcess,
						MyProcess->pr_Task.tc_Node.ln_Name);
				}
				else
				{
					Printb(data,"  Not Mounted\n");
				}
			}

			if (args[ARG_VERBOSE])
			{
				struct DateTime dt;
				struct DosAttrBuffer date={&dt.dat_Stamp,sizeof(dt.dat_Stamp)};
				char	StrDay[LEN_DATSTRING];
				char	StrDate[LEN_DATSTRING];
				char	StrTime[LEN_DATSTRING];

				dt.dat_Stamp.ds_Days = -1;
				dt.dat_Stamp.ds_Minute = -1;
				dt.dat_Stamp.ds_Tick = -1;
				if (GetDosObjectAttrTags(DOS_VOLUMENODE,dosentry,
							ADO_VN_Date,&date,
							TAG_DONE))
				{
					dt.dat_Flags   = 0; // DTF_SUBST  - Better without, imho. - Piru
					dt.dat_Format  = FORMAT_DOS;
					dt.dat_StrDay  = StrDay;
					dt.dat_StrDate = StrDate;
					dt.dat_StrTime = StrTime;
					if (DateToStr(&dt))
					{
						Printb(data,"  %s %s %s\n",
							StrTime,
							StrDay,
							StrDate);
					}
					else
					{
						Printb(data,"  <illegal date>\n");
						Printb(data,"  Days %ld Minutes %ld Ticks %ld\n",
								dt.dat_Stamp.ds_Days,
								dt.dat_Stamp.ds_Minute,
								dt.dat_Stamp.ds_Tick);
					}
				}

				if (GetDosObjectAttrTags(DOS_VOLUMENODE,dosentry,
							ADO_VN_DiskType,&value,
							TAG_DONE))
				{
					Printb(data,"  DiskType 0x%lx\n",
							value.dab_Ptr);
				}


				if (GetDosObjectAttrTags(DOS_VOLUMENODE,dosentry,
							ADO_VN_LockList,&value,
							TAG_DONE))
				{
					struct FileLock *lock;
					lock = (struct FileLock*) BADDR(value.dab_Ptr);

					while (lock)
					{
						char buf[256];
						if (NameFromLock(MKBADDR(lock),buf,sizeof(buf)-1))
						{
							buf[sizeof(buf)-1] = '\0';
							Printb(data,"    Lock 0x%lx <%s>\n",
								lock,
								buf);
						}
						else
						{
							Printb(data,"    Lock 0x%lx <Not Resolvable>\n",
								lock);
						}
						lock =(struct FileLock*) BADDR(lock->fl_Link);
					}
				}

			}

			if(SetSignal(0L,SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C)
			{
				error = RETURN_FAIL;
				SetIoErr(ERROR_BREAK);
				break;
			}
		}
		UnLockDosList(LDF_VOLUMES | LDF_READ);
		DumpPrintb(data, Output());
	}
	return(error);
}

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

int ShowAssigns(struct data	*data)
{
struct DosList	*dosentry;
int		error=RETURN_OK;

	InitPrintb(data);

	if ((dosentry=AttemptLockDosList(LDF_ASSIGNS | LDF_READ)))
	{
		while ((dosentry=NextDosEntry(dosentry, LDF_ASSIGNS)))
		{
			char	*namearg;
			char	namebuf[256];
			char	assignnamebuf[256];
			struct DosAttrBuffer name={namebuf,sizeof(namebuf)};
			struct DosAttrBuffer value={0,sizeof(ULONG)};
			struct DosAttrBuffer assignname={assignnamebuf,sizeof(assignnamebuf)};
			struct AssignList *assign;
			char buf[256];

			if ((namearg=(char*) args[ARG_NAME]))
			{
				if (GetDosObjectAttrTags(DOS_ASSIGNNODE,dosentry,
							ADO_AN_Name,&name,
							TAG_DONE))
				{
					if ((stricmp(name.dab_Ptr,namearg)!=0))
					{
						continue;
					}
				}
				else
				{
					continue;
				}
			}

			if (args[ARG_ADDRESS])
			{
				if (*((ULONG*) args[ARG_ADDRESS]) != (ULONG) dosentry)
				{
					continue;
				}
			}

			if (GetDosObjectAttrTags(DOS_ASSIGNNODE,dosentry,
						ADO_AN_Name,&name,
						TAG_DONE))
			{
				Printb(data,"<%s> AssignNode 0x%lx\n",
					name.dab_Ptr,
					dosentry);
			}

			value.dab_Ptr = NULL;
			if (GetDosObjectAttrTags(DOS_ASSIGNNODE,dosentry,
						ADO_AN_AssignName,&assignname,
						TAG_DONE))
			{
				if (assignname.dab_Ptr==NULL)
				{
					Printb(data,"<%s> AssignNode 0x%lx\n",
						name.dab_Ptr,
						dosentry);

					if (GetDosObjectAttrTags(DOS_ASSIGNNODE,dosentry,
								ADO_AN_MsgPort,&value,
								TAG_DONE))
					{
						struct Process *MyProcess;
						MyProcess = GetPortProc(data, value.dab_Ptr);
						if (MyProcess)
						{
							Printb(data,"  Process 0x%lx <%s>\n",
								MyProcess,
								MyProcess->pr_Task.tc_Node.ln_Name);
						}
						else
						{
							Printb(data,"  Not Mounted\n");
						}
					}

					if (GetDosObjectAttrTags(DOS_ASSIGNNODE,dosentry,
								ADO_AN_Lock,&value,
								TAG_DONE))
					{
						if (NameFromLock((BPTR) value.dab_Ptr,buf,sizeof(buf)-1))
						{
							buf[sizeof(buf)-1] = '\0';
							Printb(data,"  Lock 0x%lx <%s>\n",
								value.dab_Ptr,
								buf);
						}
						else
						{
							Printb(data,"    Lock 0x%lx <Not Resolvable>\n",
								value.dab_Ptr);
						}
					}
					else
					{
						Printb(data,"    Lock argument failed\n");
					}
				}
				else
				{
					Printb(data,"<%s> AssignNode 0x%lx defered to <%s>\n",
						name.dab_Ptr,
						dosentry,
						assignname.dab_Ptr);
				}
			}

			if (args[ARG_VERBOSE])
			{
				if (GetDosObjectAttrTags(DOS_ASSIGNNODE,dosentry,
							ADO_AN_Type,&value,
							TAG_DONE))
				{
					Printb(data,"  Type %ld\n",
						value.dab_Ptr);
				}
			}

			if (GetDosObjectAttrTags(DOS_ASSIGNNODE,dosentry,
						ADO_AN_AssignList,&value,
						TAG_DONE))
			{
				assign =(struct AssignList*) value.dab_Ptr;
			}
			else
			{
				assign =(struct AssignList*) NULL;
			}
			#if 0
			Printb(data,"  AssignList 0x%lx\n",
				assign);
			#endif
			while (assign)
			{
				struct FileLock *lock;
				if (NameFromLock(assign->al_Lock,buf,sizeof(buf)-1))
				{
					buf[sizeof(buf)-1] = '\0';
					Printb(data,"  + Lock 0x%lx <%s>\n",
						assign->al_Lock,
						buf);
				}
				else
				{
					Printb(data,"  + Lock 0x%lx <Not Resolvable>\n",
						assign->al_Lock);
				}

				lock = (struct FileLock*) BADDR(assign->al_Lock);
				if (args[ARG_VERBOSE])
				{
					struct DosList	*dosentry2;
					struct Process *MyProcess;

					Printb(data,"    Link 0x%lx Key 0x%lx Access %ld\n",
						lock,
						lock->fl_Link,
						lock->fl_Key,
						lock->fl_Access);

					MyProcess = GetPortProc(data, lock->fl_Task);

					if (MyProcess)
					{
						Printb(data,"    Process 0x%lx <%s>\n",
							MyProcess,
							MyProcess->pr_Task.tc_Node.ln_Name);
					}
					else
					{
						Printb(data,"    No Filesystem MsgPort\n");
					}

					dosentry2 =(struct DosList *) BADDR(lock->fl_Volume);
					if (dosentry2)
					{
						if (GetDosObjectAttrTags(DOS_VOLUMENODE,dosentry,
									ADO_VN_Name,&name,
									TAG_DONE))
						{
							Printb(data,"<%s> VolumeNode 0x%lx\n",
								name.dab_Ptr,
								dosentry);
						}
						else
						{
							Printb(data,"Lock 0x%lx fl_Volume 0x%lx is no VolumeNode\n",lock,dosentry2);
						}
					}

				}

				assign = assign->al_Next;
			}

			if(SetSignal(0L,SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C)
			{
				error = RETURN_FAIL;
				SetIoErr(ERROR_BREAK);
				break;
			}
		}
		UnLockDosList(LDF_ASSIGNS | LDF_READ);
		DumpPrintb(data, Output());
	}
	return(error);
}

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

static
void InitPrintb(struct data *data)
{
  NewList((struct List *) &printblist);
}

static
APTR printbputchproc(APTR putchdata, UBYTE ch)
{
  struct data *data = putchdata;
  struct fmtbuffer *node;

  if (ch == '\0')
    return putchdata;

  node = (APTR) RemTail((struct List *) &printblist);
  if (!node || node->index >= PRINTBBUFSIZE)
  {
    if (node)
      AddTail((struct List *) &printblist, (struct Node *) node);

    node = AllocMem(sizeof(*node) + PRINTBBUFSIZE, MEMF_ANY);
    if (!node)
      return putchdata;

    node->index = 0;
  }

  node->buf[node->index] = ch;
  node->index++;

  AddTail((struct List *) &printblist, (struct Node *) node);

  return putchdata;
}

#include <exec/rawfmt.h>
static
void Printb(struct data *data, const char *fmt, ...)
{
  va_list ap;

  va_start(ap, fmt);
//VNewRawDoFmt(fmt, RAWFMTFUNC_SERIAL, (APTR) data, ap);
  VNewRawDoFmt(fmt, printbputchproc, (APTR) data, ap);
  va_end(ap);
}

static
void DumpPrintb(struct data *data, BPTR fh)
{
  struct fmtbuffer *node;

  Flush(fh);

  while ((node = (APTR) RemHead((struct List *) &printblist)))
  {
    (void) Write(fh, node->buf, node->index);
    FreeMem(node, sizeof(*node) + PRINTBBUFSIZE);
  }

  Flush(fh);
}

